from abc import ABCMeta, abstractmethod
from io import BytesIO

import pandas as pd

from irspack.dataset.downloader import SingleZipDownloader


class BaseMovieLenstDataLoader(SingleZipDownloader):
    def _read_as_istream(self, path: str) -> BytesIO:
        bytes_array = self.zf.read(path)
        return BytesIO(bytes_array)

    @abstractmethod
    def read_interaction(self) -> pd.DataFrame:
        raise NotImplementedError("Not implemented")  # pragma: no cover



import urllib.request
from gzip import GzipFile
from pathlib import Path
from tempfile import NamedTemporaryFile
from zipfile import ZIP_DEFLATED, ZipFile

import pandas as pd

from .downloader import BaseDownloader


class AmazonMusicDataManager(BaseDownloader):
    DEFAULT_PATH = Path("~/.amazon-music.zip").expanduser()

    def _save_to_zippath(self, path: Path) -> None:
        ratings_url = "http://snap.stanford.edu/data/amazon/productGraph/categoryFiles/ratings_Digital_Music.csv"
        meta_url = "http://snap.stanford.edu/data/amazon/productGraph/categoryFiles/meta_Digital_Music.json.gz"
        with ZipFile(path, "w", compression=ZIP_DEFLATED) as zf:
            print("download rating...")
            with NamedTemporaryFile("w") as rating_temp:
                urllib.request.urlretrieve(ratings_url, rating_temp.name)
                rating_temp.seek(0)
                zf.write(rating_temp.name, "amazon-music/ratings.csv")

            print("download item meta data...")
            with NamedTemporaryFile("w") as meta_temp:
                urllib.request.urlretrieve(meta_url, meta_temp.name)
                meta_temp.seek(0)
                with GzipFile(meta_temp.name, "rb") as ifs:
                    zf.writestr("amazon-music/music-meta.json", ifs.read())

    def read_interaction(self) -> pd.DataFrame:
        return pd.read_csv(
            self._read_as_istream("amazon-music/ratings.csv"),
            header=None,
            names=["user_id", "music_id", "rating", "timestamp"],
        )

    def read_metadata(self) -> pd.DataFrame:
        results = []
        for l in self._read_as_istream("amazon-music/music-meta.json"):
            meta = eval(l.decode("utf-8"))
            results.append(
                (
                    meta["asin"],
                    meta.pop("title", None),
                    meta.pop("price", None),
                    meta.pop("categories", [None])[0],
                )
            )
        return pd.DataFrame(
            results, columns=["music_id", "title", "price", "categories"]
        ).set_index("music_id")



import os
import urllib.request
from abc import ABCMeta, abstractmethod
from io import BytesIO
from pathlib import Path
from typing import Optional, Union
from zipfile import ZipFile


class BaseDownloader(metaclass=ABCMeta):
    DEFAULT_PATH: Path
    _zf: Optional[ZipFile]

    @abstractmethod
    def _save_to_zippath(self, path: Path) -> None:
        raise NotImplementedError()

    def __init__(
        self, zippath: Optional[Union[Path, str]] = None, force_download: bool = False
    ):
        if zippath is None:
            zippath = self.DEFAULT_PATH
            if not zippath.exists():
                if not force_download:
                    download = (
                        input(
                            "Could not find {}.\nCan I download and save it there?[y/N]".format(
                                zippath
                            )
                        ).lower()
                        == "y"
                    )
                else:
                    download = True
                if download:
                    print("start download...")
                    self._save_to_zippath(zippath)
                    print("complete")
                else:
                    raise RuntimeError("could not read zipFile")

        self.zf = ZipFile(zippath)

    def _read_as_istream(self, path: str) -> BytesIO:
        bytes_array = self.zf.read(path)
        return BytesIO(bytes_array)


class SingleZipDownloader(BaseDownloader):
    DOWNLOAD_URL: str

    def _save_to_zippath(self, path: Path) -> None:
        urllib.request.urlretrieve(self.DOWNLOAD_URL, path)



from pathlib import Path

import pandas as pd

from .downloader import SingleZipDownloader


class CiteULikeADataManager(SingleZipDownloader):
    r"""Data manager for CiteULike-a dataset, which has been originally used
    in the paper "Collaborative Topic Regression with Social Regularization".
    """
    DOWNLOAD_URL = (
        "https://github.com/tohtsky/citeulike-a/archive/refs/heads/master.zip"
    )
    DEFAULT_PATH = Path("~/.citeulike-a.zip").expanduser()

    def read_interaction(self) -> pd.DataFrame:
        ifs = self._read_as_istream("citeulike-a-master/users.dat")
        result = []
        cnt = 0
        for user_id, l in enumerate(ifs):
            cnt += 1
            user_cnt: int
            n_items = 0
            for i, id_ in enumerate(l.split()):
                if i == 0:
                    user_cnt = int(id_)
                else:
                    n_items += 1
                    result.append((user_id, int(id_)))
            assert user_cnt == n_items
        return pd.DataFrame(result, columns=["user_id", "item_id"])

    def read_item_meta(self) -> pd.DataFrame:
        item_meta = pd.read_csv(
            self._read_as_istream("citeulike-a-master/raw-data.csv"), encoding="latin-1"
        )
        item_meta.loc[:, "doc.id"] -= 1
        return item_meta.set_index("doc.id")
